#!/usr/bin/env python3
# pairwise.py - Pairwise/Condorcet calculator for Python
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abiflib import *
import argparse
import json
import pathlib
from pprint import pprint
import re
import sys
import urllib.parse


def pairwise_count_dict(abifmodel):
    '''Convert abifmodel into pairwise matrix of vote counts'''
    candidates = abifmodel['candidates']
    votelines = abifmodel['votelines']

    candtoks = list(candidates.keys())

    # Initialize the return value matrix
    retval = {}
    for atok in candtoks:
        retval[atok] = {}
        for btok in candtoks:
            if atok == btok:
                retval[atok][btok] = None
            else:
                retval[atok][btok] = 0

    # Now add voteline qtys for each higher ranked cand
    for i, line in enumerate(votelines):
        thisqty = line['qty']
        lineprefs = line['prefs']
        maxrank = sys.maxsize
        for atok in candtoks:
            for btok in candtoks:
                if atok in lineprefs:
                    arank = lineprefs[atok].get('rank')
                else:
                    arank = maxrank
                if btok in lineprefs:
                    brank = lineprefs[btok].get('rank')
                else:
                    brank = maxrank
                # note that we're just ignoring arank > brank, since
                # the larger loop is only responsible for adding votes
                # when atok has a higher rank (lower number) than btok
                if atok == btok:
                    retval[atok][btok] = None
                elif arank is None or brank is None:
                    # FIXME: this condition was kludged in to make this run
                    retval[atok][btok] = 0
                elif arank < brank:
                    retval[atok][btok] += thisqty
    return retval


def winlosstie_dict_from_pairdict(candidates, pairdict):
    candtoks = list(candidates.keys())
    winlosstie_dict = {}
    for atok in candtoks:
        winlosstie_dict[atok] = {'wins': 0,
                                 'losses': 0,
                                 'ties': 0}
    for atok in candtoks:
        for btok in candtoks:
            a2b = pairdict[atok][btok]
            b2a = pairdict[btok][atok]
            # When atok == btok, that's the diagonal stripe in the
            # middle of the matrix with no values because candidates
            # don't run against each other.
            if atok != btok:
                if a2b > b2a:
                    winlosstie_dict[atok]['wins'] += 1
                    winlosstie_dict[btok]['losses'] += 1
                elif a2b == b2a:
                    winlosstie_dict[atok]['ties'] += 1
                    winlosstie_dict[btok]['ties'] += 1
                else:
                    # Avoiding counting each matchup twice by only paying
                    # attention to half of the matrix
                    pass

    stuples = sorted(winlosstie_dict.items(),
                     key=lambda item: item[1]['wins'], reverse=True)
    sorted_dict = {k: v for k, v in stuples}

    return sorted_dict


def full_copecount_from_abifmodel(abifmodel):
    '''Consolidate pairwise tally and win-loss-tie structs'''
    copecount = {}
    copecount['winningvotes'] = pairwise_count_dict(abifmodel)
    copecount['winlosstie'] = winlosstie_dict_from_pairdict(
        abifmodel['candidates'],
        copecount['winningvotes'])
    return copecount


def calc_Copeland_scores(copecount):
    point_totals = {}
    for candidate, results in copecount['winlosstie'].items():
        points = results['wins'] + (results['ties'] * 0.5)
        point_totals[candidate] = points
    sorted_point_tuples = sorted(point_totals.items(),
                                 key=lambda x: x[1],
                                 reverse=True)
    return sorted_point_tuples


def get_Copeland_winners(copecount):
    """Return a list of candidates having the highest Copeland score"""
    copescores = calc_Copeland_scores(copecount)
    winning_score = max(cscore for name, cscore in copescores)
    return [name for name, score in copescores if score == winning_score]


def Copeland_report(canddict, copecount):
    retval = ""
    #retval += f"{canddict=}\n"
    copescores = calc_Copeland_scores(copecount)
    #retval += f"odlWinner: {copescores[0][0]=} {copescores[0][1]=}\n"
    retval += f"Copeland Winner: {canddict[copescores[0][0]]} (score: {copescores[0][1]})\n"
    return retval


def main():
    """Create pairwise matrix"""
    parser = argparse.ArgumentParser(
        description='Condorcet calc')
    parser.add_argument('input_file', help='Input .abif')

    args = parser.parse_args()

    abiftext = pathlib.Path(args.input_file).read_text()
    jabmod = convert_abif_to_jabmod(abiftext)

    outstr = ""
    outstr += "======= ABIF FILE =======\n\n"
    outstr += headerfy_text_file(abiftext,
                                 filename=args.input_file)

    outstr += "\n======= PAIRWISE RESULTS =======\n\n"

    pairdict = pairwise_count_dict(jabmod)

    outstr += textgrid_for_2D_dict(
        twodimdict=pairdict,
        tablelabel='   Loser ->\nv Winner')

    print(outstr)



if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# textoutput.py - Utility functions for structured data
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abiflib import *
from abiflib.pairwise import *
import argparse
import json
import re
import sys
try:
    import texttable
    from texttable import Texttable
except:
    pass
import urllib.parse


def textgrid_for_2D_dict(twodimdict,
                         tablelabel="YYYYYYX"):
    # The first level of dict keys becomes row labels
    # The second level of dict keys becomes column labels

    retval = ""

    table = Texttable()
    ctok = list(twodimdict.keys())
    table.add_row([tablelabel] + ctok)

    for cand, inner_dict in twodimdict.items():
        try:
            mylist = list(inner_dict.values())
            table.add_row([cand] + mylist)
        except AttributeError:
            print(inner_dict)
            raise

    tabletext = table.draw()
    tabletextarray = tabletext.splitlines()

    # kludge to put "====" instead of "---" between header row and body
    header_break = tabletextarray[3]
    equalsign_break = re.sub("-", "=", header_break)
    tabletextarray[3] = equalsign_break

    subarray = []
    for i, ln in enumerate(tabletextarray):
        newln, count = re.subn(r"\|", r"+", ln, count=2)
        if count > 1:
            tabletextarray[i] = newln

    retval += "\n".join(tabletextarray)
    retval += "\n\n"
    return retval


def texttable_pairwise_and_winlosstie(abifmodel):
    def wltstr(cand):
        retval=f"{wltdict[cand]['wins']}" + "-"
        retval+=f"{wltdict[cand]['losses']}" + "-"
        retval+=f"{wltdict[cand]['ties']}"
        return retval

    pairdict = pairwise_count_dict(abifmodel)
    wltdict = winlosstie_dict_from_pairdict(abifmodel['candidates'], pairdict)
    tablelabel='   Loser ->\nv Winner'
    retval = ""

    table = Texttable()
    ctok = list(pairdict.keys())


    candkeys = sorted(pairdict.keys(),
                      key=lambda x: wltdict[x]['wins'],
                      reverse=True)

    toprow = [tablelabel] + candkeys
    table.add_row(toprow)

    dbg = False
    for ck in candkeys:
        inner_dict = pairdict[ck]
        try:
            mykeys = list(inner_dict.keys())
            myvals = list(inner_dict.values())
            invals = []
            for ik in candkeys:
                invals.append(pairdict[ck][ik])
            #candkeys = sorted(pairdict.keys(),
            #          key=lambda x: wltdict[x]['wins'],
            #          reverse=True)


            if dbg:
                print(f"{mykeys=}")
                print(f"{candkeys=}")
                print(f"{myvals=}")
                print(f"{invals=}")
            rowlabel=f"{ck} ({wltstr(ck)})"
            #table.add_row([rowlabel] + myvals)
            table.add_row([rowlabel] + invals)
        except AttributeError:
            print(inner_dict)
            raise

    tabletext = table.draw()
    tabletextarray = tabletext.splitlines()

    # kludge to put "====" instead of "---" between header row and body
    header_break = tabletextarray[3]
    equalsign_break = re.sub("-", "=", header_break)
    tabletextarray[3] = equalsign_break

    subarray = []
    for i, ln in enumerate(tabletextarray):
        newln, count = re.subn(r"\|", r"+", ln, count=2)
        if count > 1:
            tabletextarray[i] = newln

    retval += "\n".join(tabletextarray)
    retval += "\n\n"
    return retval


def headerfy_text_file(filetext, filename="???"):
    retval = ""
    retval += "============================\n"
    retval += f"{filename}\n"
    retval += "-------------------------\n"
    retval += filetext
    return retval


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# abiflib/abifregexp.py - Global ABIF-related variables and funcs
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

COMMENT_REGEX = r'''
    ^                       # beginning of line
    (?P<beforesep>[^\#]*)   # before the comment separator
    (?P<comsep>\#+)         # # or ## comment separator
    (?P<whitespace>\s*)     # optional whitespace
    (?P<aftersep>.*)        # after the # separator/whitespace
    $                       # end of line
    '''

VOTERID_REGEX = r'''
    (?P<vidprefix>\#\#VID:) # portion to match exactly to co-opt comment
    (?P<voterid>\S+)        # voter id (with no spaces)
'''

METADATA_REGEX = r'''
    ^\{                     # abif metadata lines always start with '{'
    \s*                     # whitespace
    [\'\"]?                 # optional quotation marks (single or double)
    ([\w\s]+)               # METADATA KEY
    \s*                     # moar whitespace!!!!
    [\'\"]?                 # ending quotation
    \s*                     # abif loves whitespace!!!!!
    :                       # COLON! Very important!
    \s*                     # moar whitesapce!!!1!
    [\'\"]?                 # abif also loves optional quotes
    ([\w\s\,\.\(\)\-\?\:\'/]+)  # METADATA VALUE
    \s*                     # more whitespace 'cuz
    [\'\"]?                 # moar quotes
    \s*                     # spaces the finals frontiers
    \}                      # look!  squirrel!!!!!
    $'''

CANDLINE_REGEX = r'''
    ^\=                     # the first character of candlines: "="
    \s*                     # whitespace
    ["\[]?                  # optional '[' or '"' prior to candtoken
    ([^:\"\]]*)             # candtoken; disallowed: " or ] or :
    ["\[]?                  # optional '[' or '"' after candtoken
    :                       # separator
    \[?                     # optional '[' prior to canddesc
    ([^\]]*)                # canddesc
    \]?                     # optional ']' after canddesc
    $                       # That's all, folks!
    '''

VOTELINE_REGEX = r'^(\d+):(.*)$'

VOTELINE_PREFPART_REGEX = r'''
    ^                         # start of string
    \s*                       # Optional whitespace
    (?P<candplusrate>         # <candplusrate> begin
    (
    (?P<candbare>[A-Za-z0-9_\-]*)   # <cand> (bare token)
    |
    [\"\[]                    # beginning quotation or square bracket
    (?P<candsqr>[^\"\]]*)     # <cand> (within quotes or square brackets)
    [\"\]]                    # ending quotation or square bracket
    )
    (/                        # optional slashrating begin
    (?P<rating>\d+)           # optional <rating> (number)
    \s*)?                     # optional slashrating end
    )                         # <candplusrate> end
    (?P<restofline>.*)        # the <restofline>
    $                         # end of string
    '''



#!/usr/bin/env python3
from abiflib import *
import re

'''
# ALTERNATIVE NAME 1: Bob Kiss
# ALTERNATIVE NAME 2: Andy Montroll
# ALTERNATIVE NAME 3: James Simpson
# ALTERNATIVE NAME 4: Dan Smith
# ALTERNATIVE NAME 5: Kurt Wright
# ALTERNATIVE NAME 6: Write-In
840: 5,{1,2,3,4,6}
355: 1,2,{3,4,5,6}
326: 1,{2,3,4,5,6}
'''

def _extract_alt_names_from_preflib_str(preflib_str):
    lines = preflib_str.splitlines()

    alt_names = []
    regex = r"^# ALTERNATIVE NAME (\d+): (.*)$"
    for line in lines:
        if match := re.search(regex, line):
            alt_names.append(match.group(2))

    return alt_names

def _extract_rankings_from_preflib_str(preflib_str, shortnames=None):
    # 840: 5,{1,2,3,4,6}
    # [1'BobKiss', 2'AndyMontroll', 3'JamesSimpson',
    #  4'DanSmith', 5'KurtWright', 6'Write-In']
    # 840: KurtWright>BobKiss=AndyMontrol=JamesSimpson=DanSmith=Write-In

    lines = preflib_str.splitlines()

    retval = ''
    regex = r"^(\d+): (.*)"

    numopt = len(shortnames)
    for line in lines:
        if match := re.search(regex, line):
            qty = match.group(1)
            rankpart = match.group(2)
            i=0
            tierarray = []
            for part in re.split(r'({[^}]*}|,)', rankpart):
                if not part or part == ',' or part.isspace():
                    continue
                candnums = []
                tierstr = ""
                for j, subpart in enumerate(re.finditer(r'(\d+)', part)):
                    candnums.append(subpart.group(1))
                tiertoks = [shortnames[int(n) - 1] for n in candnums]
                tierstr += "=".join(tiertoks)
                tierarray.append(tierstr)

                i += 1

            rankstr = ">".join(tierarray)
            retval += f'{qty}:{rankstr}\n'

    return retval

def convert_preflib_str_to_abif(preflib_str):
    retval = ""
    alt_names = _extract_alt_names_from_preflib_str(preflib_str)
    shortnames = []
    for o in alt_names:
        optname = re.sub(' ', '', o)
        shortnames.append(optname)
    for i, o in enumerate(alt_names):
        retval += f'={shortnames[i]}:[{o}]\n'
    retval += f'# ---------------\n'

    retval += str(
        _extract_rankings_from_preflib_str(preflib_str,
                                           shortnames=shortnames)
                  )
 
    return retval


def main():
    """Convert PrefLib file to .abif file"""
    parser = argparse.ArgumentParser(description=main.__doc__)
    parser.add_argument('input_file', help='Input PrefLib tally file')

    args = parser.parse_args()

    preflib_str = ""
    with open(args.input_file) as f:
        preflib_str += f.read()

    outstr = ""
    outstr += convert_preflib_str_to_abif(preflib_str)

    print(outstr)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from abiflib import *
import argparse
from copy import deepcopy
import pathlib
from pprint import pprint, pformat
import sys


def _eliminate_cands_from_votelines(candlist, votelines):
    '''Returns copy of votelines without candidate(s) in candlist'''
    retval = deepcopy(votelines)
    for cand in candlist:
        for vln in retval:
            if cand in vln['prefs']:
                del vln['prefs'][cand]
    return retval


def _discard_toprank_overvotes(votelines):
    retval = deepcopy(votelines)
    overvotes = 0
    for i, vln in enumerate(retval):
        prefs = vln['prefs']
        rlist = sorted(prefs.keys(), key=lambda key: prefs[key]['rank'])
        if len(rlist) > 0:
            x = 0
            xtok = rlist[x]
            xrank = prefs[xtok]['rank']
            y = len(rlist)
            yrank = 9999999999  # close enough to infinity
            # this technique should find the index of the last candidate
            # in the rlist array that has the same rank as the first
            # candidate in rlist.
            while yrank > xrank:
                y += -1
                ytok = rlist[y]
                yrank = prefs[ytok]['rank']
            # if x < y, this means there is two or more elements in the
            # rlist array with the same rank
            if x == y:
                rcand = rlist[x]
            else:
                overvotes += vln['qty']
                del retval[i]
        else:
            rcand = prefs

    return (overvotes, retval)


def _get_valid_topcand_qty(voteline):
    prefs = voteline['prefs']
    rlist = sorted(prefs.keys(), key=lambda key: prefs[key]['rank'])

    if len(rlist) > 0:
        x = 0
        xtok = rlist[x]
        xrank = prefs[xtok]['rank']
        y = len(rlist)
        yrank = 9999999999  # close enough to infinity
        # this technique should find the index of the last candidate
        # in the rlist array that has the same rank as the first
        # candidate in rlist.
        while yrank > xrank:
            y += -1
            ytok = rlist[y]
            yrank = prefs[ytok]['rank']
        # if x < y, this means there is two or more elements in the
        # rlist array with the same rank
        if x == y:
            rcand = rlist[x]
        else:
            klist = rlist[y+1:]
            nextprefs = {k: prefs[k] for k in klist if k in prefs}
            nextvln = {
                'qty': voteline['qty'],
                'prefs': nextprefs
            }
            (rcand, rqty) = _get_valid_topcand_qty(nextvln)
    else:
        rcand = prefs
    rqty = voteline['qty']
    return (rcand, rqty)


def _irv_count_internal(candlist, votelines, rounds=None, roundmeta=None):
    """
    IRV count of given votelines

    This returns the following triple:
    * winner - the winner(s) of this round, if there are any
    * rounds - round-by-round votecounts
    * roundmeta - metadata associated with all rounds
    """
    abiflib_test_log("1. looking for all_eliminated in roundmeta:")
    abiflib_test_log(pformat(roundmeta))
    if rounds is None:
        rounds = []
    if roundmeta is None:
        roundmeta = []
    roundcount = {cand: 0 for cand in candlist}

    # initializing mymeta, which will eventually be appended to roundmeta
    mymeta = {}
    mymeta['exhaustedqty'] = 0
    mymeta['overvoteqty'] = 0
    mymeta['ballotcount'] = 0
    mymeta['starting_cands'] = candlist
    mymeta['startingqty'] = sum(vln['qty'] for vln in votelines)

    (ov, prunedvlns) = _discard_toprank_overvotes(votelines)
    mymeta['overvoteqty'] += ov
    for (i, vln) in enumerate(prunedvlns):
        # (rcand, rqty, overvote) = _get_valid_topcand_qty(vln)
        (rcand, rqty) = _get_valid_topcand_qty(vln)

        mymeta['ballotcount'] += rqty
        if rcand:
            roundcount[rcand] += rqty
        else:
            mymeta['exhaustedqty'] += rqty

    total_votes = sum(roundcount.values())
    mymeta['countedqty'] = total_votes - mymeta['exhaustedqty']
    winner = None
    rounds.append(roundcount)
    roundmeta.append(mymeta)
    min_votes = min(roundcount.values())
    max_votes = max(roundcount.values())
    roundmeta[-1]['top_voteqty'] = min(roundcount.values())
    roundmeta[-1]['bottom_voteqty'] = max(roundcount.values())
    bottomcands = [c for c, v in roundcount.items() if v <= min_votes]
    if "all_eliminated" not in roundmeta[-1]:
        roundmeta[-1]['all_eliminated'] = set()
    if len(roundmeta) > 1:
        roundmeta[-1]['all_eliminated'].update(roundmeta[-2]['all_eliminated'])

    if (len(roundmeta) > 1):
        roundmeta[-1]['all_eliminated'].update(
            list(roundmeta[-2]['eliminated']))
        try:
            roundmeta[-1]['all_eliminated'].update(
                list(roundmeta[-2]['eliminated']))
        except TypeError:
            print(roundmeta[-2]['eliminated'])
            print(pformat(roundmeta))
            sys.exit()
    try:
        roundmeta[-1]['all_eliminated'].update(bottomcands)
    except TypeError:
        print(f"{bottomcands=}")
        sys.exit()
    if min_votes == max_votes:
        # This should be reached only if there's a tie between candidates
        winner = [c for c, v in roundcount.items() if v == max_votes]
        retval = (winner, rounds, roundmeta)
        roundmeta[-1]['winner'] = winner
        roundmeta[-1]['eliminated'] = set(mymeta['starting_cands']) - set(winner)
    elif max_votes > total_votes / 2:
        # This is the normal end of the IRV elimination cycle
        winner = [c for c, v in roundcount.items() if v == max_votes]
        retval = (winner, rounds, roundmeta)
        roundmeta[-1]['winner'] = winner
        roundmeta[-1]['eliminated'] = set(mymeta['starting_cands']) - set(winner)
    else:
        # We need another round, hence recursion
        roundmeta[-1]['eliminated'] = bottomcands
        nextcands = list(set(candlist) - set(bottomcands))
        nextvotelines = \
            _eliminate_cands_from_votelines(bottomcands, prunedvlns)
        (winner, nextrounds, nextmeta) = \
            _irv_count_internal(nextcands,
                                nextvotelines,
                                rounds=rounds,
                                roundmeta=roundmeta)
        retval = (winner, rounds, roundmeta)
    return retval


def IRV_dict_from_jabmod(jabmod):
    retval = {}
    canddict = retval['canddict'] = jabmod['candidates']
    candlist = list(jabmod['candidates'].keys())
    votelines = deepcopy(jabmod['votelines'])
    (retval['winner'], retval['rounds'], retval['roundmeta']) = \
        _irv_count_internal(candlist, votelines)

    winner = retval['winner']
    if len(winner) > 1:
        winnerstr = " and ".join(canddict[w] for w in sorted(winner))
    else:
        winnerstr = canddict[winner[0]]
    retval['winnerstr'] = winnerstr

    abiflib_test_log('IRV_dict_from_jabmod retval:')
    abiflib_test_log(pformat(retval))
    return retval


def get_IRV_report(IRV_dict):
    winner = IRV_dict['winner']
    rounds = IRV_dict['rounds']
    canddict = IRV_dict['canddict']
    output = ""

    for round_num, round_results in enumerate(rounds):
        thisroundmeta = IRV_dict['roundmeta'][round_num]
        eliminated = thisroundmeta.get('eliminated')
        starting_cands_str = ", ".join(
            sorted(thisroundmeta.get('starting_cands')))
        output += f"\nRound {round_num + 1}:\n"
        output += f"  Starting cands: {starting_cands_str}\n"
        output += f"  Total starting votes: {thisroundmeta['startingqty']}\n"
        output += f"  Exhausted votes: {thisroundmeta['exhaustedqty']}\n"
        output += f"  Overvotes: {thisroundmeta['overvoteqty']}\n"
        output += f"  Total counted votes: {thisroundmeta['countedqty']}\n"
        output += f"  Votes by candidate:\n"
        for candidate, votes in round_results.items():
            output += f"    {candidate}: {votes}\n"
        output += f"  Eliminated this round: {', '.join(sorted(eliminated))}\n"

    if len(winner) > 1:
        output += f"The IRV winners are {' and '.join(sorted(winner))}\n"
    else:
        output += f"The IRV winner is {winner[0]}\n"
    return output


def main():
    parser = argparse.ArgumentParser(description='IRV calculator')
    parser.add_argument('input_file', help='Input .abif')
    parser.add_argument('-j', '--json', action="store_true",
                        help='Provide raw json output')

    args = parser.parse_args()
    abiftext = pathlib.Path(args.input_file).read_text()
    jabmod = convert_abif_to_jabmod(abiftext)
    IRV_dict = IRV_dict_from_jabmod(jabmod)
    output = ""
    if args.json:
        output += json.dumps(clean_dict(IRV_dict), indent=4)
    else:
        output += candlist_text_from_abif(jabmod)
        output += get_IRV_report(IRV_dict)
    print(output)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# widjfuncs.py - funcitons for working with Electowidget (.widj) files
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

PRUNED_WIDJ_FIELDS = [
    "display_parameters", "display_results",
    "display_ballots", "allow_voting",
    "ballot_type", "max_rating", "min_rating",
    "count_subpage_ballots", "count_inline_ballots",
    "election_methods", "inline_ballot_type",
    "candidates", "inline_ballots"
]


def convert_widj_to_jabmod(widgetjson):
    """Converts electowidget JSON (widj) to the JSON ABIF model (jabmod)."""

    abifmodel = {}
    abifmodel["metadata"] = {}
    abifmodel["metadata"]["version"] = ABIF_VERSION
    abifmodel["candidates"] = {}

    # Fill in abifmodel["metadata"]
    for field in widgetjson:
        if field not in PRUNED_WIDJ_FIELDS:
            abifmodel["metadata"][field] = widgetjson[field]

    # Fill in abifmodel["candidates"]
    for candtoken, candidate_info in widgetjson["candidates"].items():
        abifmodel["candidates"][candtoken] = candidate_info["display_name"]

    # Fill in abifmodel["votelines"]
    abifmodel["votelines"] = _map_widj_ballots(widgetjson)

    return abifmodel


def _map_widj_ballots(widgetjson):
    """Maps widj ballots to abif voteline objects"""

    abif_votelines = []
    for ballot in widgetjson["inline_ballots"]:
        abif_ballot = {
            "qty": ballot["qty"],
            "prefs": {},
        }
        for candtoken, rating in ballot["vote"].items():
            abif_ballot["prefs"][candtoken] = {'rating': rating}
        abif_votelines.append(abif_ballot)

    return abif_votelines



#!/usr/bin/env python3
# textoutput.py - Utility functions for structured data
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abiflib import *
from abiflib.pairwise import *
from abiflib.html_output import *
import argparse
import json
import sys
try:
    from bs4 import BeautifulSoup
except:
    pass

def htmltable_pairwise_and_winlosstie(abifmodel,
                                      add_desc = True,
                                      snippet = False,
                                      validate = False,
                                      clean = False,
                                      modlimit = 50,
                                      svg_text = None,
                                      modifiers = set()):
    '''Generate HTML summary of election as abifmodel

    The "abifmodel" is the internal data structure for abiflib,
    referred to as "jabmod" when expressed as JSON.  This function
    accepts an abifmodel, and returns HTML that tallies pairwise
    elections for the rankings expressed in the abifmodel.
    '''

    # A few functions used within htmltable_pairwise_and_winlosstie
    def wltstr(cand):
        '''String representation of the wins, losses, and ties'''
        retval = f"{wltdict[cand]['wins']}" + "-"
        retval += f"{wltdict[cand]['losses']}" + "-"
        retval += f"{wltdict[cand]['ties']}"
        return retval


    def get_winlosstie_sorted_keys(pairdict, wltdict):
        '''Sort the candidates by win/loss/tie record'''
        # TODO, consider win-loss-tie, not just "wins"
        return sorted(pairdict.keys(),
                      key=lambda x: wltdict[x]['wins'],
                      reverse=True)

    # Initialization of key variables
    if validate:
        validate_abifmodel(abifmodel)
    retval = ""
    pairdict = pairwise_count_dict(abifmodel)
    wltdict = winlosstie_dict_from_pairdict(abifmodel['candidates'], pairdict)
    try:
        mytitle = get_title_for_html(abifmodel)
    except:
        global_namespace = globals()
        global_functions = [name for name, obj in global_namespace.items() if callable(obj)]
        print(global_functions) 
        raise
    candtoks = get_winlosstie_sorted_keys(pairdict, wltdict)
    candnames = abifmodel.get('candidates', None)

    # Soup initialization
    soup = BeautifulSoup('', 'html.parser')
    html_doc = soup.new_tag('html')

    if not snippet:
        # Soup head
        head = soup.new_tag('head')
        title = soup.new_tag('title')
        title.string = get_title_for_html(abifmodel)
        head.append(title)

        # Soup body
        body = soup.new_tag('body')
        h1 = soup.new_tag('h1')
        h1.string = get_title_for_html(abifmodel)
        body.append(h1)
    if add_desc:
        desc = soup.new_tag('p')
        desc.string = f'{get_abif_desc(abifmodel)}'
        if snippet:
            soup.append(desc)
        else:
            body.append(desc)

    # Soup table
    table = soup.new_tag('table')
    table['border'] = "1"

    # Soup table header row init
    header_text_array = ['Candidate']
    header_text_array.extend(candtoks)

    wltcolspan = len(candtoks) + 1
    candnames = abifmodel.get('candidates', None)
    has_ties_or_cycles = False

    # Soup table data rows
    # ck = column key
    # rk = row key
    for i, ck in enumerate(candtoks):
        isPastDivider = False
        candrow = soup.new_tag('tr')
        candrow_label = soup.new_tag('th')
        candrow_label.string = f"{candnames[ck]}"
        if ck != candnames[ck]:
            candrow_label.string += f" [\"{ck}\"]"
        candrow.append(candrow_label)
        candrow_wlt = soup.new_tag('td', attrs={'style': 'padding-right: 3em;'})
        candrow_wlt['colspan'] = wltcolspan
        candrow_wlt.string = f"({wltstr(ck)})"
        wincaption = \
            soup.new_tag('td', attrs={'style': 'text-align: center;'})
        if wltdict[ck]['wins'] > 0:
            appendme = soup.new_tag('div')
            appendme.string = f"{ck} victories"
            wincaption.append(appendme)
            #appendme = soup.new_tag('div',
            #                        style="float: center; white-space: nowrap;")
            #appendme.string = "victories"
            #wincaption.append(appendme)
            appendme = soup.new_tag('div', style="float: center;")
            appendme.string = "↓"
            wincaption.append(appendme)
        else:
            candrow_wlt['colspan'] += 1

        candrow.append(candrow_wlt)
        if wltdict[ck]['wins'] > 0:
            candrow.append(wincaption)

        for j, rk in enumerate(reversed(candtoks)):
            thiscell = soup.new_tag('td', style='justify-content: center;')
            # breaking out of rendering this line if we hit
            # the blank diagonal line of the matrix where
            # candidates are matched against themselves.
            if ck == rk:
                pass
            elif candtoks.index(ck) < candtoks.index(rk):
                pass
            else:
                rkscore = pairdict[rk][ck]
                ckscore = pairdict[ck][rk]
                scorespan = \
                    soup.new_tag('div', style='text-align: right;')
                winspan = \
                    soup.new_tag('div', style='white-space: nowrap;')
                winspan.string = f"{rk}: {rkscore}"
                scorespan.append(winspan)
                #breakspan = soup.new_tag('div')
                #breakspan.string = " — "
                #thiscell.append(breakspan)
                lossspan = soup.new_tag('div', style ='white-space: nowrap;')
                lossspan.string = f"{ck}: {ckscore}"
                if not rkscore > ckscore:
                    dagspan = soup.new_tag('sup')
                    dagspan.string = "†"
                    has_ties_or_cycles = True
                    lossspan.append(dagspan)
                scorespan.append(lossspan)
                thiscell.append(scorespan)
                candrow.append(thiscell)
        candrow_loss_point = soup.new_tag('td')
        if wltdict[ck]['losses'] > 0:
            candrow_loss_point.string = f"← {ck} losses"
        else:
            candrow_loss_point.string = f"{ck} is undefeated"
        candrow.append(candrow_loss_point)

        wltcolspan += -1
        table.append(candrow)

    results_div = soup.new_tag("div")
    if svg_text:
        svg_diagram = BeautifulSoup(svg_text, 'xml')
        svg_scroll = soup.new_tag("div",
                                  attrs={"class": "hscroll"})
        svg_scroll.append(svg_diagram)
        results_div.append(svg_scroll)
    table_scroll = soup.new_tag("div",
                                attrs={"class": "hscroll"})
    table_scroll.append(table)
    results_div.append(table_scroll)
    if has_ties_or_cycles:
        results_div.append(dagspan)
        results_div.append("\"Victories\" and \"losses\" sometimes aren't " +
                           "displayed in the expected location when there " +
                           "are ties and/or cycles in the results, but the " +
                           "numbers provided should be accurate.")
    if snippet:
        soup.append(results_div)
    else:
        body.append(results_div)
        html_doc.append(head)
        html_doc.append(body)
        soup.append(html_doc)

    retval += soup.prettify()

    return retval


def main():
    """Convert jabmod to html-formatted tally"""
    parser = argparse.ArgumentParser(description=main.__doc__)
    parser.add_argument('input_file', help='Input jabmod')

    args = parser.parse_args()

    with open(args.input_file, "r") as f:
        abifmodel = json.load(f)

    abifmodel['metadata']['filename'] = args.input_file
    outstr = ""
    outstr += htmltable_pairwise_and_winlosstie(abifmodel)
    print(outstr)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# textoutput.py - Utility functions for structured data
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abiflib import *
from abiflib.html_output import *

def get_abif_title(abifmodel):
    '''Title (or filename) from the abifmodel metadata'''
    metadata = abifmodel.get('metadata', None)
    defdesc = "(.metadata.title and .metadata.filename missing)"
    if metadata:
        retval = metadata.get('title', None)
        if not retval:
            retval = metadata.get('filename', defdesc)
    else:
        retval = defdesc
    return retval


def get_abif_desc(abifmodel):
    '''Description of the election from the abifmodel metadata'''
    metadata = abifmodel.get('metadata', None)
    cand_list_str = ', '.join(abifmodel['candidates'].values())
    default_desc = f"Candidate matchups for {cand_list_str}"
    if metadata:
        retval = metadata.get('description', default_desc)
    else:
        retval = default_desc
    return retval


def get_title_for_html(abifmodel):
    retval = f"abiflib/html_output.py Results: {get_abif_title(abifmodel)}"
    return retval


def validate_abifmodel(abifmodel):
    global ABIF_MODEL_LIMIT
    modlimit = ABIF_MODEL_LIMIT
    modsize = sys.getsizeof(abifmodel)
    err = f"Model size {modsize} exceeds modlimit {modlimit}"
    if modsize > modlimit:
        raise Exception(err)





#!/usr/bin/env python3
from abiflib import *
import base64
from hashlib import sha1
import re


def _extract_option_names_from_tally_sheet(tally_sheet):
    lines = tally_sheet.splitlines()

    option_names = []
    regex = r"Option (\d+).*:\s+(.*)"
    for line in lines:
        if mg := re.search(regex, line):
            option_names.append(mg.group(2))

    return option_names


def _extract_vline_rankings_from_tally_sheet(tally_sheet):
    lines = tally_sheet.splitlines()

    votelines = []
    regex = r"^V: ([-\d]+)\s+(\S+).*$"
    for line in lines:
        if mg := re.search(regex, line):
            voterid = mg.group(2)
            rankarray = [ voterid ]
            rankarray.extend(list(mg.group(1)))
            votelines.extend([rankarray])
        #else:
        #    voterid = None
        #abiflib_test_log(f"LINE: {line}")
        #abiflib_test_log(f"{votelines=}")
        #abiflib_test_log(f"{voterid=}")
    return votelines


def _trunc_sha_str(basestr, trunclen=4):
    checksum = sha1(basestr.encode('utf-8')).digest()[:trunclen]
    checksum_encoded = base64.b32encode(checksum).decode('utf-8')
    return checksum_encoded[:trunclen]


def _short_token(longstring, max_length=20, add_sha1=False):
    if len(longstring) <= max_length and \
       re.match(r'^[A-Za-z0-9]+$', longstring):
        return longstring

    wordlist = []
    for word in longstring.split():
        wordlist.append(re.sub('[^A-Za-z0-9]+', '', word))

    if add_sha1:
        base_string = ''.join(wordlist)[:max_length-4]
        retval = base_string + _trunc_sha_str(longstring)
    else:
        retval = ''.join(wordlist)[:max_length]
    return retval


def _get_short_option_names(option_names):
    retval = []
    tokset = set()
    for o in option_names:
        o2 = re.sub('(?i)none of the above', 'NOTA', o)
        optname = _short_token(o2)
        if optname in tokset:
            optname = _short_token(o2, add_sha1=True)
        tokset.add(optname)
        retval.append(optname)
    return retval


def convert_debtally_to_abif(debtallysheet, metadata={}):
    retval = ""
    option_names = _extract_option_names_from_tally_sheet(debtallysheet)

    for k in metadata.keys():
        retval += '{' + f'"{k}": "{metadata[k]}"' + '}\n'
    short_option_names = _get_short_option_names(option_names)

    for i, o in enumerate(option_names):
        retval += f'={short_option_names[i]}:[{o}]\n'
    retval += f'# ---------------\n'

    votelist = _extract_vline_rankings_from_tally_sheet(debtallysheet)

    numopt = len(option_names)
    for vote in votelist:
        tiers = []
        voterid = vote[0]
        pref = vote[1:]
        for r in range(numopt):
            strrank = str(r + 1)
            crs = [i for i, ir in enumerate(pref) if ir == strrank]
            if (crs):
                names = [short_option_names[i] for i in crs]
                tiers.append(f'{"=".join(names)}')
        tierstr = ">".join(tiers)
        retval += f'1:{tierstr}'
        retval += f"  ##VID:{voterid}\n"
    return retval


def main():
    """Convert Debian tally sheet to crude .abif file"""
    parser = argparse.ArgumentParser(description=main.__doc__)
    parser.add_argument('input_file', help='Input Debian tally file')

    args = parser.parse_args()

    debtallysheet = ""
    with open(args.input_file) as f:
        debtallysheet += f.read()

    outstr = ""
    outstr += convert_debtally_to_abif(debtallysheet)

    print(outstr)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# abiflib/core.py - core ABIF<=>jabmod conversion functions
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abiflib import *
from pprint import pprint
import copy
import inspect
import json
import os
import re
import sys
import urllib.parse

ABIF_VERSION = "0.1"
ABIF_MODEL_LIMIT = 2500

# "DEBUGARRAY" is a generic array to put debug strings in to get
# printed if there's an exception or other debugging situations
DEBUGARRAY = []

class LogfileSingleton:
    """Either append msgs to ABIFLIB_LOG or quietly munch"""
    _instance = None
    _filename = None
    _filehandle = None
    devtoolmsgs = []

    def __new__(cls, force_log=False):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            # If ABIFLIB_LOG isn't set, silently eat all msgs
            if filename := os.getenv('ABIFLIB_LOG'):
                cls._filename = os.getenv('ABIFLIB_LOG')
                if os.path.isabs(filename):
                    logdir = os.path.dirname(filename)
                else:
                    err = f"ABIFLIB_LOG needs to be an absolute path: {filename}"
                    cls.devtoolmsgs.append(err)
                    raise ValueError(err)
                if not os.path.exists(logdir):
                    err = f"Dir in ABIFLIB_LOG path doesn't exist: {logdir}"
                    cls.devtoolmsgs.append(err)
                    raise FileNotFoundError(err)
                cls._filehandle = open(filename, "a")
                msg = f"ABIFLIB_LOG set to {filename}.  "
                msg += "Amending additional debugging output there."
                cls.devtoolmsgs.append(msg)
            elif force_log:
                msg = f"ABIFLIB_LOG isn't set, and {force_log=}"
                raise ValueError(msg)
            else:
                msg = "ABIFLIB_LOG not set, so no extra logging will be done."
                cls.devtoolmsgs.append(msg)
                return cls._instance

        return cls._instance

    def log(self, msg):
        """Log a message to the file if filehandle is set; otherwise, do nothing."""
        if self._filehandle:
            self._filehandle.write(f"{msg}")
            self._filehandle.flush()

    @classmethod
    def close_file(cls):
        """Closes the file handle if open."""
        if cls._filehandle:
            cls._filehandle.close()
            cls._filehandle = None

def abiflib_test_log(msg=None):
    """Logs to file in ABIFLIB_LOG environment variable if defined."""
    logobj = LogfileSingleton()
    logobj.log(f"{msg}\n")


def abiflib_stackfunc(index=1):
    ''' Return the calling function (as index=1) or indexed function '''
    curframe = inspect.currentframe()
    callstack = inspect.getouterframes(curframe)
    return callstack[index].function


def abiflib_callstackstr(start=1, end=-1):
    ''' Return the calling stack leading to this function '''
    curframe = inspect.currentframe()
    callstack = inspect.getouterframes(curframe)
    stackstr = " < ".join([f.function for f in callstack[start:end]])
    return stackstr


def main():
    """Dev tools for abif (e.g. logging)"""
    parser = argparse.ArgumentParser(
        description='Dev tools for abif (e.g. logging)')
    args = parser.parse_args()

    teststr = "hello werld"
    devobj = LogfileSingleton(force_log=True, verbose=True)
    devobj.log(teststr)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# textoutput.py - Utility functions for structured data
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from abiflib import *
from abiflib.pairwise import *
from abiflib.html_output_common import *
from abiflib.html_output_pairwise import *
from abiflib.html_output_scorestar import *





#!/usr/bin/env python3
# abiflib/core.py - core ABIF<=>jabmod conversion functions
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abiflib import *
import copy
import json
from pprint import pprint
import re
import sys
import urllib.parse

def convert_text_to_abif(fromfmt, inputstr, cleanws=False, add_ratings=False, metadata={}):

    if (fromfmt == 'abif'):
        try:
            abifmodel = convert_abif_to_jabmod(inputstr,
                                               cleanws=cleanws,
                                               add_ratings=add_ratings)
        except ABIFVotelineException as e:
            print(f"ERROR: {e.message}")
            sys.exit()
    elif (fromfmt == 'debtally'):
        rawabifstr = convert_debtally_to_abif(inputstr, metadata=metadata)
        abifmodel = convert_abif_to_jabmod(rawabifstr)
    elif (fromfmt == 'jabmod'):
        abifmodel = json.loads(inputstr)
    elif (fromfmt == 'preflib'):
        rawabifstr = convert_preflib_str_to_abif(inputstr)
        abifmodel = convert_abif_to_jabmod(rawabifstr)
    elif (fromfmt == 'widj'):
        abifmodel = convert_widj_to_jabmod(inputstr)
    else:
        raise ABIFVotelineException(value=inputstr,
                                    message=f"Cannot convert from {fromfmt} yet.")
    retval = convert_jabmod_to_abif(abifmodel, add_ratings=False)
    return retval


def candlist_text_from_abif(jabmod):
    canddict = jabmod['candidates']
    output = ""
    output += "Candidates:\n"
    for k, v in sorted(canddict.items()):
        output += f"  {k}: {v}\n"
    return output



#!/usr/bin/env python3
# abiflib/core.py - core ABIF<=>jabmod conversion functions
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abiflib import *
from pprint import pprint, pformat
import copy
import csv
import inspect
import io
import json
import os
import re
import sys
import urllib.parse

ABIF_VERSION = "0.1"
ABIF_MODEL_LIMIT = 2500


class ABIFVotelineException(Exception):
    """Raised when votelines are missing from ABIF."""

    def __init__(self, value, message="ABIFVotelineException glares at you"):
        global DEBUGARRAY
        self.value = value
        self.message = message
        self.debugarray = DEBUGARRAY
        super().__init__(self.message)


def corefunc_init(tag="unmarked"):
    '''Initialization for all abiflib/core.py functions

    This function was added as a place to aggregate logging
    functionality for all functions in abiflib/core.py

    '''
    # abiflib_test_log(f"{tag}: {abiflib_callstackstr(start=2, end=6)}")
    return {'tag': tag}


########################
# Functions for parsing/reading abif into jabmod....
#

def convert_abif_to_jabmod(inputstr,
                           cleanws=False,
                           add_ratings=False,
                           extradata=None,
                           storecomments=False):
    """Converts an .abif string to JSON/jabmod

    'jabmod' stands for 'JSON ABIF Model', and is the internal abiflib
    data structure which is used throughout abiflib.
    """
    initval = corefunc_init(tag="f01")
    global COMMENT_REGEX, METADATA_REGEX
    global CANDLINE_REGEX, VOTELINE_REGEX
    commentregexp = re.compile(COMMENT_REGEX, re.VERBOSE)
    metadataregexp = re.compile(METADATA_REGEX, re.VERBOSE)
    candlineregexp = re.compile(CANDLINE_REGEX, re.VERBOSE)
    votelineregexp = re.compile(VOTELINE_REGEX, re.VERBOSE)

    newmodel = _get_emptyish_abifmodel()

    if len(inputstr) == 0:
        msg = f'Empty ABIF string..'
        raise ABIFVotelineException(value=inputstr, message=msg)
    # 'v' is the voteline number
    v = 0
    for i, fullline in enumerate(inputstr.splitlines()):
        matchgroup = None
        linecomment = None
        cparts = None
        # if "--cleanws" flag is given, strip leading whitespace
        if cleanws:
            fullline = re.sub(r"^\s+", "", fullline)
        # Strip the comments out first
        if (match := commentregexp.match(fullline)):
            matchgroup = 'commentregexp'
            cparts = match.groupdict()
            strpdline = cparts['beforesep']
            linecomment = cparts['comsep'] + \
                cparts['whitespace'] + cparts['aftersep']
        else:
            strpdline = fullline
            linecomment = None
        abifmodel = _process_abif_comment_line(abifmodel=newmodel,
                                               linecomment=linecomment,
                                               linenum=i,
                                               storecomments=storecomments)

        # now to deal with the substance
        if (match := candlineregexp.match(strpdline)):
            matchgroup = 'candlineregexp'
            candtoken, canddesc = match.groups()
            abifmodel = _process_abif_candline(candtoken,
                                               canddesc,
                                               abifmodel,
                                               linecomment)
        elif (match := metadataregexp.match(strpdline)):
            matchgroup = 'metadataregexp'
            mkey, mvalue = match.groups()
            abifmodel = _process_abif_metadata(
                mkey, mvalue, abifmodel, linecomment)
        elif (match := votelineregexp.match(strpdline)):
            matchgroup = 'votelineregexp'
            qty, prefstr = match.groups()
            abifmodel = _process_abif_prefline(qty,
                                               prefstr,
                                               abifmodel,
                                               linecomment)
            v += 1
        else:
            matchgroup = 'empty'

    if len(abifmodel['votelines']) == 0:
        if len(inputstr) > 20:
            msg = f'No votelines in "{inputstr[:20]}...". '
        else:
            msg = f'No votelines in "{inputstr}". '
        msg += "Votelines (like '20:A>B>C') are required in valid ABIF files."
        raise ABIFVotelineException(value=inputstr, message=msg)

    # Add in Borda-ish score if ratings are not provided
    firstprefs = abifmodel['votelines'][0]['prefs']
    firstk, firstv = list(firstprefs.items())[0]
    if not firstv.get('rating') and add_ratings:
        abifmodel = _add_ratings_to_jabmod_votelines(abifmodel)

    slist = sorted(abifmodel["votelines"], key=lambda x: x['qty'],
                   reverse=True)
    abifmodel["votelines"] = slist
    if extradata:
        abiflib_test_log(f"Ignoring {extradata=}")

    return abifmodel


def _process_abif_comment_line(abifmodel=None,
                               linecomment="",
                               linenum=0,
                               storecomments=False):
    '''Store abif comments in jabmod metadata'''
    initval = corefunc_init(tag="f02")
    if not abifmodel:
        abifmodel = _get_emptyish_abifmodel()
    commenttuple = (linenum, linecomment)
    if linecomment and storecomments:
        if not 'comments' in abifmodel['metadata']:
            abifmodel['metadata']['comments'] = []
        abifmodel['metadata']['comments'].append(commenttuple)
    return abifmodel


def _process_abif_metadata(mkey, mvalue, abifmodel, linecomment=None):
    '''Simple key-value translation of metadata lines

    This function handles the metadata lines that begin with "{".
    Each metadata line should look vaguely like a line from a "JSON
    Lines" file.

    '''
    initval = corefunc_init(tag="f03")
    # Rename ballotcount that is passed in, since this tool is going
    # to recount the ballots.
    if mkey == 'ballotcount':
        mkey = 'ballotcount_abif_metadata'
    if 'metadata' in abifmodel:
        abifmodel['metadata'][mkey] = mvalue
    else:
        abifmodel['metadata'] = {mkey: mvalue}
    return abifmodel


def _process_abif_candline(candtoken, canddesc, abifmodel, linecomment=None):
    '''_process_abif_candline maps candtokens to full candidate names'''
    initval = corefunc_init(tag="f04")
    if not 'candidates' in abifmodel:
        abifmodel['candidates'] = {candtoken: canddesc}
    else:
        abifmodel['candidates'][candtoken] = canddesc
    return abifmodel


def _get_emptyish_abifmodel():
    '''Provide initialized jabmod/abifmodel'''
    initval = corefunc_init(tag="f05")
    retval = {}
    retval['candidates'] = {}
    retval['metadata'] = {}
    retval['metadata']['ballotcount'] = 0
    retval['votelines'] = []
    return retval


def _add_ranks_to_prefjab_by_rating(inprefjab):
    '''Use candidate ratings to provide rankings'''
    initval = corefunc_init(tag="f06")
    retval = inprefjab.copy()

    # Sort cands by rating (descending order)
    cands = sorted(retval,
                   key=lambda x: int(retval[x].get("rating", 0)),
                   reverse=True)
    if not inprefjab.get(cands[0]).get('rating'):
        msg = f"Invalid call to _add_ranks_to_prefjab_by_rating"
        raise ABIFVotelineException(value=inprefjab, message=msg)

    # Assign ranks
    prevrating = None
    thisrank = 0
    for i, c in enumerate(cands):
        if i == 0:
            thisrank = 1
        elif int(retval[c].get("rating", 0)) < prevrating:
            thisrank += 1
        retval[c]["rank"] = thisrank
        prevrating = int(retval[c].get("rating", 0))
    return retval


def _add_ratings_to_jabmod_votelines(inmod):
    ''' Calculate Borda-like ratings in lieu of explicit ratings '''
    initval = corefunc_init(tag="f07")
    numcands = len(inmod['candidates'])
    outmod = copy.deepcopy(inmod)
    for vl in outmod['votelines']:
        for k, v in vl['prefs'].items():
            v['rating'] = numcands - v['rank']
    return outmod


def _parse_prefstr_to_dict(prefstr, qty=0,
                           abifmodel=None, linecomment=None):
    '''Convert prefstr portion of .abif voteline to jabvoteline
    structure.'''
    initval = corefunc_init(tag="f08")
    prefs = {}
    rank = 1

    if not abifmodel:
        abifmodel = _get_emptyish_abifmodel()

    # Split the string by commas or ranking symbols
    candidates = re.split(r",|>|=", prefstr)
    delimeters = [char for char in prefstr if char in ">,="]
    if delimeters and delimeters[0] in '>=':
        rank_or_rate = "rank"
    elif delimeters and delimeters[0] == ",":
        rank_or_rate = "rate"
    else:
        rank_or_rate = "rankone"
    candkeys = []
    for (i, candpref) in enumerate(candidates):
        candpref = candpref.strip()
        if not candpref:
            continue

        parts = candpref.split("/")
        candidate = parts[0].strip("[]\"")
        candkeys.append(candidate)
        if candidate not in abifmodel['candidates']:
            abifmodel['candidates'][candidate] = candidate

        prefs[candidate] = {}
        if i < len(candidates) - 1:
            prefs[candidate]["nextdelim"] = delimeters[i]
        if rank_or_rate == "rankone":
            rank = 1
            prefs[candidate]["rank"] = rank
        elif rank_or_rate == "rank":
            prefs[candidate]["rank"] = rank
            if i < len(candidates) - 1 and delimeters[i] == ">":
                rank += 1

        if len(parts) > 1:
            prefs[candidate]["rating"] = parts[1]
    firstcandprefs = prefs.get(candkeys[0])
    if firstcandprefs.get('rating') and not firstcandprefs.get('rank'):
        prefs = _add_ranks_to_prefjab_by_rating(inprefjab=prefs)
    return {"prefs": prefs}


def _process_abif_prefline(qty, prefstr,
                           abifmodel=None, linecomment=None):
    '''Add prefline with qty to the provided abifmodel/jabmod'''
    voteridregexp = re.compile(VOTERID_REGEX, re.VERBOSE)
    voterid = None
    if linecomment is not None:
        if (match := voteridregexp.match(linecomment)):
            cparts = match.groupdict()
            voterid = cparts['voterid']

    initval = corefunc_init(tag="f09")
    if not abifmodel:
        abifmodel = _get_emptyish_abifmodel()

    try:
        abifmodel['metadata']['ballotcount'] += int(qty)
    except:
        print(f"{abifmodel['metadata']['ballotcount']=} ")
        raise
    linepair = {}
    linepair['qty'] = int(qty)
    linepair['prefs'] = \
        _parse_prefstr_to_dict(prefstr,
                               qty=qty,
                               abifmodel=abifmodel,
                               linecomment=linecomment)['prefs']
    linepair['comment'] = linecomment
    linepair['prefstr'] = prefstr.rstrip()
    if voterid is not None:
        linepair['voterid'] = voterid
    abifmodel['votelines'].append(linepair)
    return abifmodel

########################
# Functions for converting jabmod into abif....
#


def convert_jabmod_to_abif(abifmodel, add_ratings=False):
    """Converts 'jabmod' to an .abif string.

    'jabmod' stands for 'JSON ABIF Model', and is the internal abiflib
    data structure which is used throughout abiflib.
    """
    initval = corefunc_init(tag="f10")
    abif_string = ""
    abif_string += "#------- metadata -------\n"

    for field, value in abifmodel["metadata"].items():
        jstr = json.dumps(value)
        abif_string += f'{{{field}: {jstr}}}\n'

    abif_string += "#------ candlines ------\n"
    for candtoken in abifmodel["candidates"]:
        candqtoken = _abif_token_quote(candtoken)
        abif_string += f"={candqtoken}:"
        abif_string += f"[{abifmodel['candidates'][candtoken]}]\n"

    abif_string += "#------- votelines ------\n"
    for (i, jabvoteline) in enumerate(abifmodel["votelines"]):
        try:
            is_ordered = jabvoteline["orderedlist"]
        except KeyError:
            is_ordered = False

        abif_chunk = _get_votelinestr_from_jabvoteline(jabvoteline)
        abif_string += abif_chunk

    return abif_string


def _abif_token_quote(candtoken):
    '''Add square brackets to a candidate token if necessary'''
    initval = corefunc_init(tag="f11")
    quotedcand = urllib.parse.quote_plus(candtoken)
    if quotedcand == candtoken:
        candtoken = candtoken
    else:
        candtoken = f"[{candtoken}]"
    return candtoken


def ranklist_from_jabmod_voteline(voteline):
    """Construct list of candtoks in order of ranking"""
    initval = corefunc_init(tag="f12")
    orderedcands = []
    toklist = list(voteline['prefs'].keys())
    firstcand = toklist[0]
    firstrank = voteline['prefs'][firstcand].get('rank', None)
    firstrating = voteline['prefs'][firstcand].get('rating', None)
    if firstrank:
        orderedcands = toklist.copy()
        orderedcands.sort(
            key=lambda x: voteline['prefs'][x]['rank'], reverse=False)
    elif firstrating:
        orderedcands = toklist.copy()
        orderedcands.sort(
            key=lambda x: voteline['prefs'][x]['rating'], reverse=True)
    else:
        orderedcands = toklist
    return orderedcands


def _prefstr_from_ranked_line(sortedprefs):
    '''provide .abif string from prefs in jabmod form'''
    initval = corefunc_init(tag="f13")
    prefstrfromranks = ""
    rank = 1

    for name, data in sortedprefs:
        prefstrfromranks += _abif_token_quote(name)
        if 'rating' in data and data['rating'] is not None:
            prefstrfromranks += f"/{data['rating']}"
        if 'nextdelim' in data:
            if data['nextdelim'] == '>':
                rank += 1
            prefstrfromranks += data['nextdelim']
    return prefstrfromranks


def _prefstr_from_ratings(sortedprefs):
    '''provide .abif string from ratings in jabmod form'''
    initval = corefunc_init(tag="f14")
    tiered_cands = []
    current_rating = 0
    current_tier = []
    for name, data in sortedprefs:
        try:
            rating = int(data['rating'])
        except TypeError:
            rating = None

        if rating != current_rating:
            if current_tier:
                tiered_cands.append(current_tier)
            current_rating = rating
            current_tier = []
        current_tier.append({name: data})
    if current_tier:
        tiered_cands.append(current_tier)

    prefstrfromratings = ""
    for i, tierblob in enumerate(tiered_cands):
        rank = i + 1
        lastindexi = len(tiered_cands) - 1
        for j, thistier in enumerate(tierblob):
            thistiercount = len(thistier)
            for k, ckey in enumerate(thistier.keys()):
                candqtoken = _abif_token_quote(ckey)
                prefstrfromratings += candqtoken
                rating = thistier[ckey]['rating']
                prefstrfromratings += f'/{rating}'
            tierblobcount = len(tierblob) - 1
            if j < tierblobcount:
                prefstrfromratings += '='
        if i < lastindexi:
            prefstrfromratings += '>'
    return prefstrfromratings


def _get_votelinestr_from_jabvoteline(jabvoteline):
    '''Convert voteline data structure to abif string for voteline

    "jabvoteline" is the jabmod/JSON structure for a voteline
    "votelinestr" is the .abif string representation of a voteline
    '''
    initval = corefunc_init(tag="f15")
    local_abif_str = ""

    try:
        prefitems = sorted(jabvoteline['prefs'].items(),
                           key=lambda item: (-int(item[1].get('rating')),
                                             item[0])
                           )
        has_full_ratings = True
    except TypeError:
        prefitems = jabvoteline['prefs'].items()
        has_full_ratings = False

    if has_full_ratings:
        prefstr = _prefstr_from_ratings(prefitems)
    else:
        prefstr = _prefstr_from_ranked_line(prefitems)

    local_abif_str += f"{jabvoteline['qty']}:{prefstr}\n"
    #abiflib_test_log(f"func13: {local_abif_str=}")
    return local_abif_str


def _modify_jabmod(jabmod, modtuple):
    """Modify jabmod structure with list of tuples

    The tuples provided provide two things:
    1) A list which represents the path to the value being modified
    2) The new value for this key/index/whatever.
    """
    def _modify_jabmod_internal(this_jabmod, path):
        if len(path) == 1:
            this_jabmod[path[0]] = new_value
        else:
            next_key = path[0]
            _modify_jabmod_internal(this_jabmod[next_key], path[1:])

    path, new_value = modtuple
    _modify_jabmod_internal(jabmod, path)


########################
#  Misc functions for data cleaning and conversion
#


def clean_dict(data):
    '''Recursive function for converting sets to lists in a dict

    This function makes is intended to make it easier to convert
    arbitrary Python datastructures to something that works for JSON
    output.  As of June 2024, it only converts sets to lists.
    '''
    if isinstance(data, set):
        return list(data)
    elif isinstance(data, dict):
        return {k: clean_dict(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [clean_dict(item) for item in data]
    else:
        return data


def get_ranking_output_csv(abifmodel):
    # Use keys as field names
    outputhandle = io.StringIO()
    fieldnames = ['voterid']
    fieldnames.extend(abifmodel['candidates'])
    csvwriter = csv.DictWriter(outputhandle, fieldnames=fieldnames)
    csvwriter.writeheader()
    i = 0
    for y, vln in enumerate(abifmodel['votelines']):
        if vln['qty'] > 0:
            for z in range(vln['qty']):
                i += 1
                rlinedict = {ckey:
                             vln['prefs'][ckey]['rank']
                             for ckey in vln['prefs'].keys()}
                if vln.get('voterid'):
                    rlinedict['voterid'] = vln['voterid']
                else:
                    rlinedict['voterid'] = f"voter{i:06d}"
                csvwriter.writerow(rlinedict)
        else:
            msg = f"Invalid voteline: {vln}\n"
            raise ABIFVotelineException(value=vln, message=msg)
    retval = outputhandle.getvalue()
    return retval


def main():
    """core functions of abiflib

    This script allows for testing of core functions of abiflib"""
    initval = corefunc_init(tag="main")
    parser = argparse.ArgumentParser(
        description='Core functions of abiflib (and some cli stuff)')
    parser.add_argument('-f', '--func',
                        help='stub function to run',
                        default='pref_to_dict',
                        choices=['pref_to_dict',
                                 'emptyish',
                                 'modify_jabmod'])
    parser.add_argument('string', help='string(s) to pass to function',
                        default=None, nargs='+')
    args = parser.parse_args()

    if args.func == 'pref_to_dict':
        if not args.string:
            parser.error("pref_to_dict requires prefline as string[0]")
        parseout = _parse_prefstr_to_dict(f"{args.string[0]}")
        print(json.dumps(parseout, indent=4))
    elif args.func == 'emptyish':
        emptyish = _get_emptyish_abifmodel()
        print(json.dumps(emptyish, indent=4))
    elif args.func == 'modify_jabmod':
        if not args.string:
            parser.error("modify_jabmod requires filename as string[0]")
        else:
            print(f"FIXME: need to load jabmod file {args.string}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# textoutput.py - Utility functions for structured data
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abiflib import *
from abiflib.scorestar import *
from abiflib.html_output import *
import argparse
import html
import json
import sys
try:
    from bs4 import BeautifulSoup
except:
    pass


def html_score_and_star(jabmod):
    retval = ""
    content = STAR_report(jabmod)
    escaped_content = [html.escape(line) for line in content]

    # FIXME: proper escaping needed for the values
    # 2024-03-31
    # I REALLY SHOULD JUST USE FLASK FOR THIS
    basicstar = STAR_result_from_abifmodel(jabmod)
    escaped_content.append(json.dumps(basicstar, indent=4))
    scaled = scaled_scores(jabmod, target_scale=50)
    escaped_content.append(json.dumps(scaled, indent=4))
    soup = BeautifulSoup('', 'html.parser')
    
    pre_tag = soup.new_tag('pre')
    pre_tag.string = ''.join(escaped_content)
    
    soup.append(pre_tag)
    retval += str(soup)
    return retval


def main():
    """Convert abif to html-formatted STAR results"""
    parser = argparse.ArgumentParser(description=main.__doc__)
    parser.add_argument('input_file', help='Input jabmod')

    args = parser.parse_args()

    with open(args.input_file, "r") as f:
        inputstr = f.read()
    abifmodel = convert_abif_to_jabmod(inputstr)

    abifmodel['metadata']['filename'] = args.input_file
    outstr = ""
    outstr += html_score_and_star(abifmodel)
    print(outstr)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
vizelect - functions for election visualizations
"""
# Copyright (C) 2024 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abiflib import *
from graphviz import Digraph
import json
import os
import sys


def copecount_diagram(copecount, outformat='svg', is_inline=False):
    winningvotes = copecount['winningvotes']
    winlosstie = copecount['winlosstie']

    # Determine the candidate with the highest Copeland score (most wins)
    top_candidate = max(winlosstie,
                        key=lambda candidate: winlosstie[candidate]['wins'])

    dot = Digraph(comment='Pairwise Matchup Visualization')

    # Add nodes with win-loss-tie counts
    for candidate, scores in winlosstie.items():
        wins = scores['wins']
        losses = scores['losses']
        ties = scores['ties']
        label = f"{candidate}"
        label += f"\n{wins}-{losses}-{ties}"
        # Include full description only for the top candidate
        if candidate == top_candidate:
            label += f"\n({wins} wins, {losses} losses, {ties} ties)"
        dot.node(candidate, label)

    # Add edges for matchups with full candidate names in labels
    for wcand, wvotes in winningvotes.items():
        for lcand, wtally in wvotes.items():
            ltally = winningvotes[lcand][wcand]
            if wcand != lcand and wtally >= ltally:
                label_text = f"← {wcand}: {wtally}\n{lcand}: {ltally}"
                dot.edge(wcand, lcand, label=label_text)

    diagram_output = dot.pipe(format=outformat).decode('utf-8')

    is_inline = True
    # strip out initial xml cruft if this intended for embedding in html
    if is_inline:
        # TODO: does it make sense to generalize this beyond svg?
        start_index = diagram_output.find('<svg')
        if start_index != -1:
            diagram_output = diagram_output[start_index:]

    return diagram_output



#!/usr/bin/env python3
# pairwise.py - Pairwise/Condorcet calculator for Python
#
# Copyright (C) 2023 Rob Lanphier
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abiflib import *
from abiflib.pairwise import *
import argparse
import json
import math
import pathlib
import re
import sys


def basic_score_result_from_abifmodel(abifmodel):
    retval = {candtok: {} for candtok in abifmodel['candidates'].keys()}
    for cand in abifmodel['candidates'].keys():
        retval[cand]['candname'] = abifmodel['candidates'][cand]
        retval[cand]['score'] = 0
        retval[cand]['votercount'] = 0
    for voteline in abifmodel['votelines']:
        qty = voteline['qty']
        for cand, candval in voteline['prefs'].items():
            rating = int(candval.get('rating', 0))
            retval[cand]['score'] += rating * qty
            if rating > 0:
                retval[cand]['votercount'] += qty
    return retval


def enhanced_score_result_from_abifmodel(abifmodel):
    retval = {}
    newscores = basic_score_result_from_abifmodel(abifmodel)
    ranklist = sorted(newscores.keys(), key=lambda x: newscores[x]["score"], reverse=True)
    retval['scores'] = newscores

    scores = {ct: newscores[ct]['score'] for ct in abifmodel['candidates'].keys()}

    # Calculate the ranks, ensuring that equal scores results in equal ranks
    rank = 0
    for i, c in enumerate(ranklist):
        if i > 0:
            pscore = newscores[ranklist[i-1]]['score']
        else:
            pscore = math.inf
        tscore = newscores[c]['score']
        if pscore > tscore:
            rank += 1
        newscores[c]['rank'] = rank
    retval['ranklist'] = ranklist
    retval['total_all_scores'] = sum(x['score']
                                     for x in retval['scores'].values())
    retval['totalvoters'] = abifmodel['metadata']['ballotcount']
    return retval


def STAR_result_from_abifmodel(abifmodel):
    retval = enhanced_score_result_from_abifmodel(abifmodel)
    bc = retval['totalvoters']
    retval['round1winners'] = retval['ranklist'][0:2]
    copecount = full_copecount_from_abifmodel(abifmodel)

    fin1 = retval['fin1'] = retval['ranklist'][0]
    fin2 = retval['fin2'] = retval['ranklist'][1]
    fin1n = retval['fin1n'] = retval['scores'][fin1]['candname']
    fin2n = retval['fin2n'] = retval['scores'][fin2]['candname']

    f1v = retval['fin1votes'] = copecount['winningvotes'][fin1][fin2]
    f2v = retval['fin2votes'] = copecount['winningvotes'][fin2][fin1]
    retval['final_abstentions'] = bc - f1v - f2v
    if f1v > f2v:
        retval['winner'] = fin1n
    elif f2v > f1v:
        retval['winner'] = fin2n
    else:
        retval['winner'] = f"tie {fin1n} and {fin2n}"
    return retval


def score_report(jabmod):
    retval = ""
    totalvoters = jabmod['metadata']['ballotcount']
    sr = enhanced_score_result_from_abifmodel(jabmod)
    for candtok in sr['ranklist']:
        candinfo = sr['scores'][candtok]
        retval += f"- {candinfo['score']} points"
        retval += f" (from {candinfo['votercount']} voters)"
        retval += f" -- {candinfo['candname']}\n"

    retval += f"Voter count: {totalvoters}\n"
    winnertok = sr['ranklist'][0]
    retval += f"Score Winner: {sr['scores'][winnertok]['candname']}\n"
    return retval


def STAR_report(jabmod):
    retval = ""
    sr = STAR_result_from_abifmodel(jabmod)
    tvot = sr['totalvoters']
    retval += f"Total voters: {tvot}\n"
    retval += f"Scores:\n"
    for candtok in sr['ranklist']:
        candinfo = sr['scores'][candtok]
        retval += f"- {candinfo['score']} stars"
        retval += f" (from {candinfo['votercount']} voters)"
        retval += f" -- {candinfo['candname']}\n"
    retval += f"Finalists: \n"
    retval += f"- {sr['fin1n']} preferred by {sr['fin1votes']} of {tvot} voters\n"
    retval += f"- {sr['fin2n']} preferred by {sr['fin2votes']} of {tvot} voters\n"
    retval += f"- {sr['final_abstentions']} abstentions\n"
    retval += f"STAR Winner: {sr['winner']}\n"
    return retval


def scaled_scores(jabmod, target_scale=100):
    retval = {}
    scores = STAR_result_from_abifmodel(jabmod)
    ballotcount = jabmod['metadata']['ballotcount']
    retval['max_rating'] = jabmod['metadata'].get('max_rating')
    retval['total_all_scores'] = scores['total_all_scores']
    scale = target_scale / retval['total_all_scores']
    retval['scale_factor'] = scale
    scaled_total = 0
    candidates = jabmod['candidates']
    retval['canddict'] = {}
    for candtoken, candname in candidates.items():
        candscore = scores['scores'][candtoken]['score']
        scaled_score = candscore * scale
        retval['canddict'][candtoken] = {
            'candname': candname,
            'scaled_score': scaled_score,
            'score': candscore
        }
        scaled_total += scaled_score
    retval['scaled_total'] = scaled_total
    return retval


def main():
    """Create score array"""
    parser = argparse.ArgumentParser(
        description='Takes abif and returns score results')
    parser.add_argument('input_file', help='Input .abif')

    args = parser.parse_args()

    abiftext = pathlib.Path(args.input_file).read_text()
    jabmod = convert_abif_to_jabmod(abiftext, add_ratings=True)

    outstr = ""
    report_abif_file = True
    report_raw_jabmod = False
    report_raw_score_json = False
    report_raw_STAR_json = False
    report_score_results = False
    report_star_results = True
    report_scaled_scores = True
    if report_abif_file:
        outstr += "======= ABIF File =======\n"
        outstr += abiftext

    if report_raw_jabmod:
        outstr += "======= ABIF Model (jabmod) =======\n"
        outstr += json.dumps(jabmod, indent=4)

    if report_raw_score_json:
        outstr += "\n======= Raw Score JSON =======\n"
        scoreres = enhanced_score_result_from_abifmodel(jabmod)
        outstr += json.dumps(scoreres, indent=4)

    if report_raw_STAR_json:
        outstr += "\n======= Raw STAR JSON =======\n"
        retval = STAR_result_from_abifmodel(jabmod)
        outstr += json.dumps(retval, indent=4)

    if report_score_results:
        outstr += "\n======= Score Results =======\n"
        outstr += score_report(jabmod)

    if report_star_results:
        outstr += "\n======= STAR Results =======\n"
        outstr += STAR_report(jabmod)

    if report_scaled_scores:
        outstr += "\n======= Scaled Scores ========\n"
        outstr += json.dumps(scaled_scores(jabmod=jabmod,
                                           target_scale=50),
                             indent=4)
    print(outstr)


if __name__ == "__main__":
    main()
